CREATE OR REPLACE FUNCTION do_rollup(queryText text, tempTableName text, siteId text, targetTable text, autoIndex boolean, partitionColumn text, dropTempTable boolean) RETURNS boolean AS $$
DECLARE
	schemaTable text[];
	allColumns text[];
	primaryColumns text[];
	alterSql text;
	dropSql text;
	columnInfo record;
	tempTableCreated boolean;
	foundIgnore boolean;
	tableExists boolean;
	sourceTableName text;
	sourceSchemaName text;
	targetTableName text;
	targetSchemaName text;
	columnList text;
	i int;
BEGIN
	-- we either select query into temporary table or take an existing one, but not both
	if queryText is not null and tempTableName is not null then
		raise exception 'Cannot specify queryText and tempTableName';
	end if;

	if queryText is not null then
		-- select results into temporary table.. if target table doesnt exist, we can simply
		-- rename, otherwise we can insert into target () select from () 
		tempTableName := 'do_rollup_' || pg_backend_pid();
		raise warning 'Selecting data into table %', tempTableName;
		execute 'create table ' || tempTableName || ' as ' || queryText;
		tempTableCreated := true;
		sourceTableName := tempTableName;
		sourceSchemaName := null;
		-- we always cleanup our temp tables if they're not moved
		dropTempTable := true;
	else
		schemaTable := regexp_split_to_array(tempTableName, E'\\.');
		if array_upper(schemaTable, 1) > 1 then
			-- we have schema and table
			sourceTableName := schemaTable[2];
			sourceSchemaName := schemaTable[1];
		else
			sourceTableName := tempTableName;
			sourceSchemaName := null;
		end if;
		tempTableCreated := true;
		raise warning 'Using existing table %, schema %', sourceTableName, sourceSchemaName;
	end if;
	-- go ahead and add siteid if there
	if siteId is not null then
		execute 'alter table ' || tempTableName || ' add column if not exists site_id text default '''|| siteId || ''';';
	end if;

	allColumns := ARRAY[]::text[];
	primaryColumns := ARRAY[]::text[];
	alterSql := '';
	dropSql := '';
	foundIgnore := false;

	for columnInfo in select column_name, udt_name  from information_schema.columns where table_name  = sourceTableName and (sourceSchemaName is null or table_schema = sourceSchemaName) order by ordinal_position asc loop
		raise warning 'Found column %, type %', columnInfo.column_name, columnInfo.udt_name;
		if columnInfo.column_name ~ 'ignore.*' then
			foundIgnore := true;
			-- no need drop columns in temp table, we simple dont select them
			--dropSql := dropSql || 'alter table ' || tempTableName || ' drop column ' || columnInfo.column_name || ';';
		else
			alterSql := alterSql || 'alter table ' || targetTable || ' add column if not exists ' || columnInfo.column_name || ' ' || columnInfo.udt_name || ';';
			allColumns := array_append(allColumns, columnInfo.column_name::text);
			-- assume all columns before first ignore are primary
			if foundIgnore is false then
				primaryColumns := array_append(primaryColumns, columnInfo.column_name::text);
			end if;
		end if;
	end loop;

	-- make sure we index on site_id
	primaryColumns := array_append(primaryColumns, 'site_id');

	raise warning 'Columns examined, add: %, drop %', alterSql, dropSql;

	-- see if current table exists or not
	schemaTable := regexp_split_to_array(targetTable, E'\\.');
	if array_upper(schemaTable, 1) > 1 then
		raise warning 'Checking for table % in schema %', schemaTable[2], schemaTable[1];
		-- we have schema and table
		targetTableName := schemaTable[2];
		targetSchemaName := schemaTable[1];
		select into tableExists count(*) = 1 from pg_tables where tablename = schemaTable[2] and schemaname = schemaTable[1];
	else
		raise warning 'Checking for table %', targetTable; 
		-- just table
		targetTableName := targetTable; 
		targetSchemaName := NULL;
		select into tableExists count(*) = 1 from pg_tables where tablename = targetTable;
	end if;

	raise warning 'Table % exists: %', targetTable, tableExists;

	if tableExists is false and dropTempTable is true then
		-- drop columns in temp table, rename, and move schema if we need to 
		dropSql := dropSql || 'alter table ' || tempTableName || ' rename to ' || targetTableName || ';';
		-- move schema if we need to
		if targetSchemaName is not null then
			if sourceSchemaName is not null then
				dropSql := dropSql || 'alter table ' || sourceSchemaName || '.' || targetTableName || ' set schema ' || targetSchemaName || ';';
			else
				dropSql := dropSql || 'alter table ' || targetTableName || ' set schema ' || targetSchemaName || ';';
			end if;
		end if;
		raise warning 'Moving table % to %, schema % via query: %', tempTableName, targetTableName, targetSchemaName, dropSql;
		execute dropSql;
	else
		if tableExists is false then
			-- get here if the table didn't exist but we didnt want to delete the source
			execute 'create table ' || targetTable || ' ()';
		end if;
		-- fixup table before delete
		execute alterSql;
		if autoIndex is true and array_upper(primaryColumns, 1) > 0 then
			-- delete duplicate keys
			columnList := array_to_string(primaryColumns, ',');
			raise warning 'Deleting duplicate keys in % that match %', targetTable, primaryColumns;
			--raise warning 'DELETE: %', 'delete from ' || targetTable || '  where (' || columnList || ') in (select ' || columnList || ' from ' || tempTableName || ')';
			 execute 'delete from ' || targetTable || '  where (' || columnList || ') in (select ' || columnList || ' from ' || tempTableName || ')';
		end if;
		columnList := array_to_string(allColumns, ',');
		-- copy data over
		alterSql := 'insert into ' || targetTable || '(' || columnList || ') select ' || columnList || ' from ' || tempTableName ||';';
		raise warning 'Target table % exists, adding new columns and copying data: %', targetTableName, alterSql;
		execute alterSql;
	end if;

	-- autoindex
	if autoIndex is true and array_upper(primaryColumns, 1) > 0 then
		raise warning 'Creating indexes on primary columns: %', primaryColumns;
		alterSql := '';
		FOR i IN 1 .. array_upper(primaryColumns, 1)
		LOOP
			alterSql := alterSql || 'create index if not exists ' ||primaryColumns[i] || '_idx on ' || targetTable || '('||primaryColumns[i]||');';
		END LOOP;
		execute alterSql;
	end if;

	-- fix permissions
	raise warning 'Setting up permissions';
	execute 'grant SELECT on ' || targetTable || '  to reports';

	-- if partitioning is on, create partition if need to, otherwise make sure data is partitioned
	if partitionColumn is not null then
		if tableExists is false then 
			-- just support time partitioning for now.. could expose more options
			alterSql := 'alter table ' || targetTable || ' alter column ' || partitionColumn || ' set not null;' || 'SELECT partman.create_parent(''' || targetTable || ''', ''' || partitionColumn || ''', ''time'', ''daily'');';
			raise warning 'Initializing partitioning on column %: %', partitionColumn, alterSql;
		else
			alterSql := '';
			raise warning 'Updating partitioning on column %: %', partitionColumn, alterSql;
		end if;
		alterSql := alterSql || 'select partman.partition_data_time(''' || targetTable || ''');';
		execute alterSql;
	end if;

	-- cleanup
	if tableExists is true and tempTableCreated is true and dropTempTable is true then
		execute 'drop table ' || tempTableName;
	end if;

	return true;
END;
$$ LANGUAGE plpgsql;
